// ..................................................................
//
//                             mdApps
//
//              Copyright  1998 by Martin Djerns
//
// ..................................................................
//
// Usage :
// This unit shall support me with small "native" windows applications
// ..................................................................
// History
// 1th April 1998 : Initial date
// + TmdApp      (A standard program with a window handle)
// + TmdTrayApp  (A tray program with popup menu)
// 2th April 1998:
// + Added ProcessMessage(s)
// ..................................................................
// Description
//
// TmdApp:
// This is the base application, and creates a window handle
// and make sure that all messages is send to the objects in
// real Delphi style, which means that you can pick up any message
// with the message directive
// Ex.
//   Procedure WMMyMessage(Var Msg : TMessage); MW_MYMESSAGE;
// but also commands is postet to the objects.
// NOTE: A program can ONLY contain ONE TmdApp (or descant) class
// since the StdWndProc do not support more than one application.
//
// TmdTrayApp:
// This is a Tray application, which offers a icon in the tray,
// a functions for adding menu items and functins for the three
// "standard" tray functions.
// NOTE: AddMenuItem adds items to the menu from the buttom, so
// the first added will be the bottom one.
// Ex.
//   AddMenuItem('&Close',miClose,[]);        // Close menu item
//   AddmenuItem('-',0,[]);                   // Separator
//   AddmenuItem('&Show',miShow,[msDefault]); // Show menu item
// results in :
//   Show
//   ----
//   Close
// ..................................................................
// Hope that you will create a nice program by using this!
// ..................................................................

unit mdApps;

interface  

Uses
  Windows,
  Messages,
  ShellApi;

Const
  smdApp  = 'mdApp';

Type

// ..................................................................
// TmdApp
// ..................................................................

  TmdApp = class(TObject)
  Private
    FHandle : THandle;
    FTerminated : Boolean;
  Protected
    Function MakeWindow(Const Name : String):Boolean; Virtual;
    Procedure FreeWindow; Virtual;
    Procedure WMCommand(Var Msg : TMessage); Message WM_COMMAND;
  Public
    Constructor Create(Const Name : String); Virtual;
    Destructor Destroy; Override;

    Procedure Run; Virtual;
    Procedure Terminate; Virtual;
    Procedure Close; Virtual;

    Function ProcessMessage : Boolean;
    Procedure ProcessMessages;

    Property Handle : THandle Read FHandle;
    Property Terminated : Boolean Read FTerminated;
  end;


// ..................................................................
// TmdTrayApp
// ..................................................................
Const
  WM_TRAY = WM_USER + 999;

Type
  TMenuState = Set of (msChecked, msDefault, msDisabled);
  TTipString = String[63]; // Since a tip maximum can be 63 chars long
  TWMTrayMessage = Record  // Windows message from the tray...
    Msg: Cardinal;
    NotUsed,
    Command,
    Result: Longint;
  end;

  TmdTrayApp = class(TmdApp)
  Private
    FMenuHandle : HMenu;
    FIconHandle : HIcon;
    FIconData : TNotifyIconData;
    FTip : TTipString;
  Protected
    Procedure MakeTray; Virtual;
    Procedure UpdateTray; Virtual;
    Procedure RemoveTray; Virtual;

    Procedure TrayMsg(Var Msg : TWMTrayMessage); Message WM_TRAY;
    Procedure OnShowMenu; Virtual;
    Procedure OnSimple; Virtual;
    Procedure OnDefault; Virtual;
  Public
    Constructor Create(Const Name : String); Override;
    Destructor Destroy; Override;

    Procedure AddMenuItem(Const Caption : String; ID : Word; State : TMenuState); Virtual;
    Procedure LoadDefaultIcon; Virtual;

    Property MenuHandle : HMenu Read FMenuHandle;
    Property IconHandle : HIcon Read FIconHandle Write FIconHandle;
    Property IconData : TNotifyIconData Read FIconData Write FIconData;
    Property Tip : TTipString Read FTip Write FTip;
  end;

implementation

// ..................................................................
Var
  InternalApp : TmdApp = NIL;

Function StdWndProc(Window: HWND; Message, WParam: Longint;
  LParam: Longint): Longint; stdcall;
Var
  Msg : TMessage;
Begin
  Msg.Msg := Message;
  Msg.wParam := WParam;
  Msg.lParam := LParam;
  InternalApp.Dispatch(Msg);
  Result := Msg.Result;
end;

Procedure StrLCopy(Dest, Src : PChar; Max : Integer);
Var
  Cnt : Integer;
Begin
  Cnt := 0;
  While ((Cnt < Max) AND (Src[Cnt] <> #0)) do
  Begin
    Dest[Cnt] := Src[Cnt];
    Inc(Cnt);
  end;
  Dest[Cnt] := #0;
end;

// ..................................................................

Constructor TmdApp.Create(Const Name : String);
Begin
  If InternalApp = NIL Then
  Begin
    Inherited Create;
    InternalApp := Self;
    If NOT MakeWindow(Name) Then
      InternalApp := NIL;
  end;
end;

Destructor TmdApp.Destroy;
Begin
  FreeWindow;
  Inherited Destroy;
end;

Function TmdApp.MakeWindow(Const Name : String):Boolean;
Var
  TempClass : TWndClass;
Begin
  TempClass.style := 0;
  TempClass.lpfnWndProc := @StdWndProc;
  TempClass.cbClsExtra := 0;
  TempClass.cbWndExtra := 0;
  TempClass.hInstance := HInstance;
  TempClass.hIcon := 0;
  TempClass.hCursor := 0;
  TempClass.hbrBackground := 0;
  TempClass.lpszMenuName := NIL;
  TempClass.lpszClassName := smdApp;

  Result := RegisterClass(TempClass) <> 0;
  If Result Then
  Begin
    FHandle := CreateWindow(smdApp,PChar(Name),0,0,0,0,0,0,0,HInstance,NIL);
    SetWindowLong(FHandle, GWL_WNDPROC, LongInt(@StdWndProc));
  end;
end;

Procedure TmdApp.FreeWindow;
Begin
end;

Function TmdApp.ProcessMessage : Boolean;
var
  Msg: TMsg;
Begin
  Result := False;
  If PeekMessage(Msg, 0, 0, 0, PM_REMOVE) then
  Begin
    Result := True;
    If Msg.Message <> WM_QUIT then
    Begin
      TranslateMessage(Msg);
      DispatchMessage(Msg);
    end
    else
      FTerminated := True;
  end;
end;

Procedure TmdApp.Run;
Begin
  If FHandle = 0 Then
    Exit;
  While NOT FTerminated do
  Begin
    If NOT ProcessMessage Then
      WaitMessage;
  end;
end;

Procedure TmdApp.ProcessMessages;
Begin
  While ProcessMessage do
    ;
end;

Procedure TmdApp.Terminate;
Begin
  PostQuitMessage(0);
end;

Procedure TmdApp.Close;
Begin
  Terminate;
end;

Procedure TmdApp.WMCommand(Var Msg : TMessage);
Begin
  Msg.Msg := Msg.wParam;
  Dispatch(Msg);
end;

// ..................................................................

Constructor TmdTrayApp.Create(Const Name : String);
Begin
  Inherited Create(Name);
  FMenuHandle := CreatePopupMenu;
  FTip := Name;
  MakeTray;
end;

Destructor TmdTrayApp.Destroy;
Begin
  RemoveTray;
  DestroyMenu(FMenuHandle);
  Inherited Destroy;
end;

Procedure TmdTrayApp.AddMenuItem(Const Caption : String; ID : Word; State : TMenuState);
Var
  MI : TMenuItemInfo;
Begin
  MI.cbSize := SizeOf(TMenuItemInfo);
  MI.fMask  := MIIM_CHECKMARKS or MIIM_DATA or MIIM_ID or
               MIIM_STATE or MIIM_SUBMENU or MIIM_TYPE;
  If (Caption = '-') OR (Caption = '') Then
    MI.fType := MFT_SEPARATOR
  else
    MI.fType := MFT_STRING;

  MI.fState := 0;
  If msChecked IN State Then
    MI.fState := MI.fState OR MFS_CHECKED;
  If msDefault IN State Then
    MI.fState := MI.fState OR MFS_DEFAULT;
  If msDisabled IN State Then
    MI.fState := MI.fState OR MFS_DISABLED;

  MI.wID := ID;
  MI.hSubMenu := 0;
  MI.hBmpChecked := 0;
  MI.hBmpUnchecked := 0;
  MI.dwTypeData := PChar(Caption);
  InsertMenuItem(MenuHandle, 0, True, MI);
end;

Procedure TmdTrayApp.MakeTray;
Begin
  FIconData.cbSize := SizeOf(TNotifyIconData);
  FIconData.wnd    := Handle;
  FIconData.uID    := 1;
  FIconData.uCallBackMessage := WM_TRAY;
  If IconHandle = 0 Then
    LoadDefaultIcon;
  FIconData.hIcon  := IconHandle;
  StrLCopy(FIconData.szTip, PChar(String(FTip)),64);  // Internal function
  FIconData.uFlags := nif_Message or nif_Icon or nif_Tip;
  Shell_NotifyIcon (NIM_ADD, @IconData);
end;

Procedure TmdTrayApp.RemoveTray;
Begin
  FIconData.uFlags := 0;
  Shell_NotifyIcon (NIM_DELETE, @FIconData);
end;

Procedure TmdTrayApp.UpdateTray;
Begin
  FIconData.hIcon  := IconHandle;
  If FTip <> '' Then
  Begin
    StrLCopy(FIconData.szTip, PChar(String(FTip)),64);  // Internal function
    FIconData.uFlags := nif_Message or nif_Icon or nif_Tip;
  end
  else
    FIconData.uFlags := nif_Message or nif_Icon;

  Shell_NotifyIcon (NIM_MODIFY, @FIconData);
end;

Procedure TmdTrayApp.TrayMsg(Var Msg : TWMTrayMessage);
Begin
  Msg.Result := LongInt(True);
  Case Msg.Command of
    WM_RButtonDown     : OnShowMenu;
    WM_LButtonDown     : OnSimple;
    WM_LButtonDblClk   : OnDefault;
  else
    Msg.Result := LongInt(False);
  end;
end;

Procedure TmdTrayApp.OnShowMenu;
Var
  Pt : TPoint;
Begin
  // ---------------------------------
  // This is a bit strange, but I always
  // got the "hourglass" cursor the first
  // time a popup menu were shown...
  // Now I set it pr. hand to the default,
  // but another Tray proggram of mine
  // (the pressor of this class) were
  // not showing the "hourglass"!!!!!!!!!
  SetCursor(LoadCursor(0, IDC_ARROW));
  // ---------------------------------
  SetForegroundWindow(Handle);
  GetCursorPos (Pt);
  TrackPopupMenu(MenuHandle,
                 TPM_LEFTALIGN OR TPM_RIGHTBUTTON,
                 Pt.x,
                 Pt.y,
                 0,  // Must be 0
                 Handle,
                 NIL);
end;

Procedure TmdTrayApp.OnSimple;
Begin
end;

Procedure TmdTrayApp.OnDefault;
Begin
end;

Procedure TmdTrayApp.LoadDefaultIcon;
Begin
  IconHandle := LoadIcon(HInstance, 'MAINICON');
end;

end.
